// PPM tester for ATmega328 linked to 128x64 monochrome OLED display
// detects PPM signal (typically from a radio control transmitter buddy box port or similar) and displays up to 12 channels numerically and with an oscilloscope-style graphic
// by ceptimus 2016-10-14
//

// connect OLED display SCL to SCL (A5)
// connect OLED display SDA to SDA (A4)
// officially you're supposed to fit pull-up resistors (2K2 or similar) to SCL and SDA, but I found my example worked fine with no resistors

// PPM signal connected to D2.  A good way to protect the ATmega against wrong polarity or wrong level of this signal is to connect the signal through a 10K resistor and series 0.1uF capacitor
// this allows the ATmega to detect and display a PPM signal even if it has negative polarity (say swinging between -2V and -10V levels relative to the ATmega ground)

#include "SX1X06.h"

#define MAX_CHANNELS 16

// minimum sync interval in half-microsecond units. example: 6000 means that any pulse interval > 3 millseconds is recognized as sync 
#define SYNC_INTERVAL_HU 6000

volatile uint8_t ppmChannels; // number of channels detected
volatile bool ppmPolarity; // true for positive-going pulses

// timestamps for the edges of the PPM signal in half-microsecond units plus one guard-element to hold any overflow channel data
volatile uint32_t ppmEdgeHu[MAX_CHANNELS * 2 + 3]; // first element is a timestamp for the whole frame, remaining elements are time offsets relative to that first element

void setup(void) {
  // set 16-bit timer1 to run at 2MHz
  TCCR1A = 0x00;
  TCCR1B = 0xA2; // CTC with clk/8 (2MHz) input
  OCR1A = 0x0000; // generate TIMER1_COMPA interrupt every 32768 microseconds when TCNT1 wraps
  TIMSK1 = 0x02; // enable TIMER1_COMPA interrupt
  
  // enable D2 pin to generate interrupts on both edge transitions
  EICRA = 0x01; 
  EIMSK = 0x01; // enable D2 pinchange interrupts  

  // initialize oled display
  delay(200);
  SX1X06::start();
  delay(200);
  SX1X06::fillDisplay(' ');

  pinMode(13, OUTPUT); // while waiting for sync, toggle LED output so an external oscilloscope can monitor 'busy wait' periods
}

volatile uint16_t timer1_overflow_count = 0;

ISR(TIMER1_COMPA_vect) { // occurs every 32768 microseconds
  ++timer1_overflow_count; // combining this 16 bit overflow count with the 16 bit timer gives a 32-bit half-microsecond timer that wraps every 35 minutes 47.483648 seconds
}

volatile uint8_t edges = 0; // edge counter
volatile uint32_t prevHu = 0UL; // timestamp of previous edge in half-microsecond units

ISR(INT0_vect) { // occurs when pin D2 changes state (+ve or -ve edge)
  bool state = PIND & _BV(2); // high or low state of D2
  uint32_t hu = timer1_overflow_count;
  uint16_t t = TCNT1;
  if((TIFR1 & _BV(OCF1A)) && (t < 0xFFFFU)) { // if timer overflowed but interrupt overflow counter didn't happen yet
    hu++;
  }
  hu = (hu << 16) + t; // timestamp of edge transition in half-microsecond units
  uint32_t elapsedHu = hu - prevHu; // half-microseconds since previous edge
  prevHu = hu;  
  if (elapsedHu > SYNC_INTERVAL_HU) { // if sync interval has elapsed then this is the first edge of the next frame
    ppmChannels = edges > 3 ? (edges >> 1) - 1 : 0; // number of channels seen in last frame (two edges per pulse and one more pulse than there are channels)
    ppmPolarity = state; // if we just saw a positive edge at the start of the frame then the polarity must be positive and vice-versa
    ppmEdgeHu[0] = hu;
    edges = 1;
  } else { // one of the edges within the frame
    ppmEdgeHu[edges] = hu - ppmEdgeHu[0]; // half-microseconds since start of frame
    if (++edges > MAX_CHANNELS * 2 + 2) {
      edges = MAX_CHANNELS * 2 + 2; // we're only interested in the edges for MAX_CHANNELS (plus a guard register to absorb any extra channel data)
    }
  }
}

uint32_t currentHu(void) { // return current half-microsecond timestamp
  uint8_t oldSREG = SREG;
  cli();
  uint32_t hu = timer1_overflow_count;
  uint16_t t = TCNT1;
  if((TIFR1 & _BV(OCF1A)) && (t < 0xFFFFU)) {
    hu++;
  }
  SREG = oldSREG;
  return (hu << 16) + t;
}

bool toggle = false; // when busy-waiting for sync pulse use this to toggle LED output on and off
void loop(void) {
  char printBuffer[22];
  uint8_t image[2]; // 1 pixel wide by 16 pixel high image: image[0] is the top 8 pixels, image[1] is the bottom 8 pixels: least significant pixels are drawn at the top
  
  // the interrupt routines above gather all the PPM timing data - all we need to do here is analyze and display it.
  uint8_t numChannels = 0;
  uint32_t frameHu, elapsedHu; // time for complete frame in half-microsecond units and time elapsed since last start-of-frame in same units
  uint32_t edgeHu[MAX_CHANNELS * 2 + 3];
  
  // data is constantly changing as new frames arrive
  // we don't really want to display some data from one frame and some other data from the next
  // so here we attempt to capture a snapshot of a single frame - or detect that no PPM signal is currently being received
  
  if (ppmChannels) {
    uint32_t lastFrameStartHu = ppmEdgeHu[0];
    uint32_t timeout = 200000;
    while (lastFrameStartHu == ppmEdgeHu[0]) { // wait for a new frame to begin
      if (toggle) { // toggle LED on and off while waiting
        PORTB &= ~_BV(5);
        toggle = false;
      } else {
        PORTB |= _BV(5);
        toggle = true;
      }
      if (!--timeout) { // better to have a busy wait loop like this rather than use millis() or similar timer methods as they generate additional interrupts that could affect pulse measurement accuracy
        break;
      }
    }
    timeout = 200000;
    while (edges < ppmChannels * 2 + 2) { // wait for the start of the sync pulse
      if (toggle) {
        PORTB &= ~_BV(5);
        toggle = false;
      } else {
        PORTB |= _BV(5);
        toggle = true;
      }
      if (!--timeout) {
        break;
      }
    }
    numChannels = ppmChannels;
    elapsedHu = currentHu() - ppmEdgeHu[0];
    frameHu = ppmEdgeHu[0] - lastFrameStartHu;
    for (uint8_t i = edges; i; i--) { // capture snapshot
        edgeHu[i] = ppmEdgeHu[i]; 
    }
  }

  if (!numChannels || !frameHu || elapsedHu > 1000000UL) { // if no start of frame seen in previous half-second
    numChannels = 0;
    SX1X06::displayString6x8(0, 0, F("     PPM Tester      "), 1);
    for (uint8_t i = 1; i <= 7; i++) {
      if (i == 2) { 
        SX1X06::displayString6x8(i, 0, F("    By ceptimus      "), 0);
      } else if (i == 4) { 
        SX1X06::displayString6x8(i, 0, F("     No Signal       "), 0);
      } else {
        SX1X06::displayString6x8(i, 0, F("                     "), 0);
      }
    }
    // rub out one row of vertical pixels at each side of the 'scope trace' display - these are not overwritten by normal text characters
    image[0] = image[1] = 0x00;
    SX1X06::drawImage(image, 0, 6, 1, 2);
    SX1X06::drawImage(image, 127, 6, 1, 2);
  } else { // display the frame data
    // display frame length in ms with 0.001 ms resolution
    sprintf(printBuffer, "%3d", frameHu / 2000UL);
    sprintf(printBuffer + 3, ".%03d ms", (frameHu % 2000UL) >> 1);
    SX1X06::displayString6x8(0, 0, printBuffer, 0);

    // calculate pulse length average duration, using all the pulses in the frame
    uint32_t pulseWidthTotalHu = 0UL;
    uint32_t pulseStartHu = 0UL; // all timestamps in the frame, except the first, are offsets from first pulse leading edge timestamp - so first pulse start time is effectively zero
    uint8_t p = 1; // index to pulse edge timestamps array - initially pointing to trailing edge of first pulse
    for (uint8_t i = 0; i <= numChannels; i++) { // note "i <= numChannels" as the number of pulses is numChannels + 1 
      uint32_t pulseEndHu = edgeHu[p];
      pulseWidthTotalHu += edgeHu[p] - pulseStartHu;
      pulseStartHu = edgeHu[++p]; // get timestamp of next pulse leading edge
      p++; // advance index to point at trailing edge of next pulse
    }
    
    // display average pulse length in us with 1 us resolution
    sprintf(printBuffer, "%8d us", (pulseWidthTotalHu + numChannels + 1) / ((numChannels + 1) * 2));
    SX1X06::displayString6x8(0, 10, printBuffer, 0);

    // extract the channel data (time intervals between pulses) in half-microsecond units
    uint32_t channel[MAX_CHANNELS];
    
    pulseStartHu = 0UL; // all timestamps in the frame, except the first, are offsets from first pulse leading edge timestamp - so first pulse start time is effectively zero
    p = 2; // index to leading edge timestamp of next pulse
    for (uint8_t i = 0; i < numChannels; i++) {
      uint32_t nextPulseStartHu = edgeHu[p];
      channel[i] = nextPulseStartHu - pulseStartHu;      
      pulseStartHu = nextPulseStartHu; // origin is now leading edge of next pulse
      p += 2;
    }

    // display up to 12 channels on oled display in us with 1 us resolution, blanking the areas for the latter channels when numChannels < 12
    for (uint8_t i = 0; i < 12; i++) {
      if (i < numChannels) {
        sprintf(printBuffer, "%6d ", channel[i] >> 1);
      } else {
        sprintf(printBuffer, "       ");
      }
      SX1X06::displayString6x8(i / 3 + 2, (i % 3) * 7, printBuffer, 0);
    }

    // draw the 'scope trace'
    uint32_t interval = (frameHu + 64UL) >> 7; // divide frame duration by 128 (rounded) - now each horizontal pixel represents interval time period (units half-microseconds)
    // trace is drawn on bottom two character rows (16 pixels) of display
    // scan one horizontal pixel per loop so we need to send two bytes of data to the display
    // we dont use the top two pixels to provide a good separation between the trace and channels 10, 11, 12 (if present) so the trace is 14 pixels high
    
    // first draw an 7-pixel lead-in and the first edge at x = 7
    image[0] = ppmPolarity ? 0x00 : 0x04; // lead-in is low for positive going pulses and high for negative going pulses
    image[1] = ppmPolarity ? 0x80 : 0x00;
    for (uint8_t x = 0; x < 7; x++) {
      SX1X06::drawImage(image, x, 6, 1, 2);
    }
    uint32_t timeHu = interval >> 1;
    uint8_t edge = 1;
    bool state = ppmPolarity;
    bool transition = true;
    for (uint8_t x = 7; x < 128; x++) {
      while (edge < numChannels * 2 + 2 && edgeHu[edge] < timeHu) {
        state = !state;
        transition = true;
        edge++;
      }
      if (transition) {
        image[0] = 0xFC;
        image[1] = 0xFF;
      } else {
        image[0] = state ? 0x04 : 0x00; // lead-in is low for positive going pulses and high for negative going pulses
        image[1] = state ? 0x00 : 0x80;
      }
      SX1X06::drawImage(image, x, 6, 1, 2);
      timeHu += interval;
      transition = false;
    }    
  }
}

