/*  TWI (I2C)  code to drive 128x64 monochrome oled display module 
 *  ceptimus.  September 2016
 */
#include "SX1X06.h"
#include <Wire.h>

void cmd(uint8_t c) {
  Wire.beginTransmission(SX1X06_I2C_ADDRESS);
  Wire.write(0);
  Wire.write(c);
  Wire.endTransmission();
}

void SX1X06::fillDisplay(uint8_t c) { // fill whole display using character c
  for (uint8_t row = 0; row < SX1X06_CHARHEIGHT; row++) {
    for (uint8_t x = 0; x < SX1X06_LCDWIDTH; x += 6) {
      displayChar6x8(row, x, c);
    }
  }
}

void SX1X06::start(void) {
  Wire.begin();
  Wire.setClock(400000L);
  
  // initial setup commands for the display
#ifdef SSD1306  
  cmd(SX1X06_SETMULTIPLEX); cmd(0x3F);
  cmd(SX1X06_SETDISPLAYOFFSET); cmd(0x00);
  cmd(SX1X06_SETSTARTLINE);
  cmd(SX1X06_SEGREMAP | 0x01);
  cmd(SX1X06_COMSCANDEC);
  cmd(SX1X06_SETCOMPINS); cmd(0x12);
  cmd(SX1X06_SETCONTRAST); cmd(0xFF);
  cmd(SX1X06_DISPLAYALLON_RESUME);
  cmd(SX1X06_NORMALDISPLAY);
  cmd(SX1X06_SETDISPLAYCLOCKDIV); cmd(0x80);
  cmd(SX1X06_CHARGEPUMP); cmd(0x14);
  cmd(SX1X06_SETPRECHARGE); cmd(0x22);
  cmd(SX1X06_MEMORYMODE); cmd(0x01);
  cmd(SX1X06_SETVCOMDETECT); cmd(0x20);
  cmd(SX1X06_DISPLAYON);
#endif

#ifdef SH1106  
  cmd(0xAE); // display off
  cmd(0x02); // lower column address
  cmd(0x10); // higher column address
  cmd(0x40); // display start line
  cmd(0xB0); // page address
  cmd(0x81); // contrast control
  cmd(0x80); // max contrast
  cmd(0xA1); // segment remap
  cmd(0xA6); // normal/reverse
  cmd(0xA8); // multiplex ratio
  cmd(0x3F); // duty = 1/32
  cmd(0xAD); // set charge pump enable
  cmd(0x8B); // external VCC
  cmd(0x30); // VPP
  cmd(0xC8); // Com scan direction
  cmd(0xD3); // set display offset
  cmd(0x00); // = 0x20
  cmd(0xD5); // oscillator division
  cmd(0x80); // = 0x80
  cmd(0xD9); // pre-charge period
  cmd(0x1F); // = 0x1F
  cmd(0xDA); // COM pins
  cmd(0x12); // = 0x12
  cmd(0xDB); // vcomh
  cmd(0x40); // = 0x40
  cmd(0xAF); // display on
#endif
}
 
void SX1X06::displayString6x8(uint8_t row, uint8_t col, const char *s, uint8_t rvsField) {
  if (row > SX1X06_CHARHEIGHT - 1) {
    row = SX1X06_CHARHEIGHT - 1;
  }
  if (rvsField) { rvsField = 0x80; }
  
  while (uint8_t c = *s++) {
    displayChar6x8(row, SX1X06_LEFTMARGIN + 6 * col++, c ^ rvsField);
  }  
}

void SX1X06::displayString6x8(uint8_t row, uint8_t col, const __FlashStringHelper *s, uint8_t rvsField) {
  if (rvsField) { rvsField = 0x80; }
  uint8_t *p = (uint8_t PROGMEM *)s;
  while (uint8_t c = pgm_read_byte_near(p++)) {
    displayChar6x8(row, SX1X06_LEFTMARGIN + 6 * col++, c ^ rvsField);
  }
}

void SX1X06::displayChar6x8(uint8_t row, uint8_t x, uint8_t c) {
  uint8_t xorMask = c & 0x80 ? 0xFF : 0x00;  // if MSB set display character reverse field

#ifdef SSD1306
  cmd(SX1X06_PAGEADDR);
  cmd(row);
  cmd(row);
  cmd(SX1X06_COLUMNADDR);
  cmd(x);
  cmd(x < SX1X06_LCDWIDTH - 6 ? x + 5 : SX1X06_LCDWIDTH - 1);
#endif

#ifdef SH1106
  cmd(0xB0 + row);
  x += 2;
  cmd(x & 0x0F);
  cmd(((x >> 4) & 0x0F) | 0x10);
#endif

  Wire.beginTransmission(SX1X06_I2C_ADDRESS);
  Wire.write(0x40);
  Wire.write(xorMask);
 
  uint16_t p = ((c & 0x7F) - 32) * 5;
  for (uint16_t i = 1; i < 6; i++ ) {
    Wire.write(pgm_read_byte_near(font + p++) ^ xorMask);
  }
  Wire.endTransmission();
}

void SX1X06::drawImage(uint8_t *img, uint8_t x, uint8_t row, uint8_t width, uint8_t rows) {
#ifdef SSD1306
  cmd(SX1X06_PAGEADDR);
  cmd(row);
  cmd(row + 1);
  cmd(SX1X06_COLUMNADDR);
  cmd(x);
  cmd(x);
  Wire.beginTransmission(SX1X06_I2C_ADDRESS);
  Wire.write(0x40);
  for (uint8_t i = width * rows; i > 0; i--) {
    Wire.write(*img++);
  }
  Wire.endTransmission();
#endif

#ifdef SH1106
  // SH1106 (I think) doesn't do vertical addressing mode, so we have to send the data a row at a time
  x += 2;
  for (uint8_t i = 0; i < rows; i++) {
    cmd(0xB0 + i + row);
    cmd(x & 0x0F);
    cmd(((x >> 4) & 0x0F) | 0x10);
    Wire.beginTransmission(SX1X06_I2C_ADDRESS);
    Wire.write(0x40);
    for (uint8_t j = 0; j < width; j++) {
      Wire.write(img[i + j * rows]);
    }
    Wire.endTransmission();
  }
#endif
}


